from typing import Dict

from fares.dimension import resolve_slot_dimensions
from fares.subscription_offering import SubscriptionOffering


def is_subscription_offering_applicable(
    subscription_offering: SubscriptionOffering, slot: Dict
):
    rule_set = subscription_offering.applicability_rules()
    context = resolve_slot_dimensions(slot)
    return rule_set.evaluate(context)



from fares.dimension.dimension import SlotDimensions


def resolve_slot_dimensions(slot):
    return {
        "slot.od_cluster_session": (
            slot[SlotDimensions.ORIGIN_CLUSTER_ID.value],
            slot[SlotDimensions.DESTINATION_CLUSTER_ID.value],
            slot[SlotDimensions.SESSION.value],
        ),
        "slot.origin_cluster_id": slot[SlotDimensions.ORIGIN_CLUSTER_ID.value],
        "slot.destination_cluster_id": slot[
            SlotDimensions.DESTINATION_CLUSTER_ID.value
        ],
        "slot.session": slot[SlotDimensions.SESSION.value],
        "slot.route_id": slot[SlotDimensions.ROUTE_ID.value],
        "slot.start_time": slot[SlotDimensions.START_TIME.value],
    }



from dataclasses import dataclass
from datetime import datetime

from rule_engine.operator import In
from rule_engine.rule import RuleSet, Rule


@dataclass(frozen=True)
class Applicability:
    origin_cluster_id: str
    destination_cluster_id: str
    session: str


@dataclass
class SubscriptionOffering:
    id: str
    name: str
    rides: int
    currency: str
    validity_in_days: int
    is_carry_forward: bool
    activation_date: datetime
    deactivation_date: datetime
    zone_id: str
    amount: int
    subscription_type: str
    applicability: [Applicability]
    created_at: datetime

    @classmethod
    def from_json(cls, subscription_offering):
        return SubscriptionOffering(
            id=subscription_offering["id"],
            name=subscription_offering["name"],
            rides=subscription_offering["rides"],
            currency=subscription_offering["currency"],
            validity_in_days=subscription_offering["validity_in_days"],
            is_carry_forward=subscription_offering["is_carry_forward"],
            activation_date=subscription_offering["activation_date"],
            deactivation_date=subscription_offering["deactivation_date"],
            zone_id=subscription_offering["zone_id"],
            amount=subscription_offering["amount"],
            subscription_type=subscription_offering["subscription_type"],
            applicability=[
                Applicability(
                    app["origin_cluster_id"],
                    app["destination_cluster_id"],
                    app["session"],
                )
                for app in subscription_offering["applicability"]
            ],
            created_at=datetime.fromisoformat(subscription_offering["created_at"]),
        )

    def applicability_rules(self) -> RuleSet:
        return RuleSet(
            [
                Rule(
                    "slot.od_cluster_session",
                    In(),
                    [
                        (app.origin_cluster_id, app.destination_cluster_id, app.session)
                        for app in self.applicability
                    ],
                )
            ]
        )



import pytest

from fares import is_subscription_offering_applicable, SubscriptionOffering


@pytest.fixture
def subscription_offering_single_od_session():
    return {
        "id": "first_sub",
        "name": "Morning sub",
        "rides": "10",
        "currency": "INR",
        "validity_in_days": "20",
        "is_carry_forward": "True",
        "activation_date": "2019-02-28T12:56:41.589953+00:00",
        "deactivation_date": "2019-02-28T12:56:41.589953+00:00",
        "zone_id": "zone1",
        "amount": "1000",
        "subscription_type": "TRIAL_PASS",
        "applicability": [
            {
                "origin_cluster_id": "cluster_1",
                "destination_cluster_id": "cluster_2",
                "session": "MORNING",
            }
        ],
        "created_at": "2019-02-28T12:56:41.589953+00:00",
    }


@pytest.fixture
def subscription_offering_multiple_od_session():
    return {
        "id": "first_sub",
        "name": "Morning sub",
        "rides": "10",
        "currency": "INR",
        "validity_in_days": "20",
        "is_carry_forward": "True",
        "activation_date": "2019-02-28T12:56:41.589953+00:00",
        "deactivation_date": "2019-02-28T12:56:41.589953+00:00",
        "zone_id": "zone1",
        "amount": "1000",
        "subscription_type": "TRIAL_PASS",
        "applicability": [
            {
                "origin_cluster_id": "cluster_2",
                "destination_cluster_id": "cluster_1",
                "session": "MORNING",
            },
            {
                "origin_cluster_id": "cluster_2",
                "destination_cluster_id": "cluster_1",
                "session": "EVENING",
            },
        ],
        "created_at": "2019-02-28T12:56:41.589953+00:00",
    }


@pytest.fixture
def morning_slot():
    return {
        "origin_cluster_id": "cluster_1",
        "destination_cluster_id": "cluster_2",
        "session": "MORNING",
        "route_id": "route_id_1",
        "start_time": "2019-02-28T4:56:41.589953+00:00",
    }


@pytest.fixture
def evening_slot():
    return {
        "origin_cluster_id": "cluster_2",
        "destination_cluster_id": "cluster_1",
        "session": "EVENING",
        "route_id": "route_id_1",
        "start_time": "2019-02-28T12:56:41.589953+00:00",
    }


def test_is_offering_applicable_happy_single_od(
    subscription_offering_single_od_session, morning_slot
):
    assert is_subscription_offering_applicable(
        SubscriptionOffering.from_json(subscription_offering_single_od_session),
        morning_slot,
    )


def test_is_offering_applicable_happy_multiple_od(
    subscription_offering_multiple_od_session, evening_slot
):
    assert is_subscription_offering_applicable(
        SubscriptionOffering.from_json(subscription_offering_multiple_od_session),
        evening_slot,
    )


def test_is_offering_applicable_sad_multiple_od(
    subscription_offering_multiple_od_session, morning_slot
):
    assert not is_subscription_offering_applicable(
        SubscriptionOffering.from_json(subscription_offering_multiple_od_session),
        morning_slot,
    )
